import click
import whisk.setup
import logging
import os

from subprocess import PIPE, STDOUT

logger = logging.getLogger(__name__)

@click.command()
@click.option('-d', '--dir', default=os.getcwd(), help="The project directory.", show_default=True)
def cli(dir):
    """
    Sets up the project environment.
    This is called by default after `whisk create` and should be run manually after cloning an existing project.

    See :func:`whisk.setup.setup` for the full list of steps that are performed.
    """
    whisk.setup.setup(dir)



import os

class cd:
    """
    Context manager for changing the current working directory
    Based on https://stackoverflow.com/questions/431684/how-do-i-change-the-working-directory-in-python/13197763#13197763
    """
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)



#!/usr/bin/env python

"""Tests for `whisk` package."""

import pytest
import shutil
import os

# https://click.palletsprojects.com/en/7.x/testing/
from click.testing import CliRunner

from whisk import whisk
from whisk.cli import cli
from pathlib import Path

@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_cli():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    help_result = runner.invoke(cli.cli, ['--help'])
    assert help_result.exit_code == 0

def test_create_via_cli():
    """Test creating an app via the CLI."""
    project_name = "project_name_cli"
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(cli.cli,['create', project_name, '--no-setup'])
        assert result.exit_code == 0
        assert os.path.exists((Path(os.getcwd()) / project_name))

def test_create(tmpdir):
    temp = tmpdir.mkdir('whisk-project')
    project_name = "project_name"
    out_dir = Path(temp).resolve()
    whisk.create(project_name, output_dir=out_dir)
    shutil.rmtree(out_dir)



import click
from whisk.project import Project
import os
from subprocess import call
project = Project()

@click.group()
def cli():
    pass

@cli.command()
@click.option("--bucket", default="whisk-" + project.name.replace("_","-"), show_default=True)
def remote_add(bucket):
    """
    Adds a DVC S3 remote as the default remote using the provided S3 bucket.
    """
    call("dvc remote add -d s3 s3://{}/".format(bucket), shell=True)

@cli.command()
@click.option("--name", default="s3", show_default=True)
def remote_remove(name):
    """
    Removes the DVC S3 remote.
    """
    call("dvc remote remove {}".format(name), shell=True)

@cli.command()
def setup():
    """
    Initializes dvc, adds a local remote, and installs git post-checkout hooks.

    Run `dvc destroy` to revert.
    """
    if os.system("venv/bin/dvc status > /dev/null 2>&1") != 0:
        # Initializing DVC
        call("venv/bin/dvc init > /dev/null", shell=True)
        # Setting up default local DVC remote
        call("venv/bin/dvc remote add -d local /tmp/dvc-storage", shell=True)
    if not os.path.isfile(".git/hooks/post-checkout"):
        # Installing Git hooks into the DVC repository
        call("venv/bin/dvc install > /dev/null", shell=True)



import os
import sys
import pytest
import shutil
from pathlib import Path
from cookiecutter import main
import whisk
from whisk.whisk import create
import whisk.setup

args = {
        'project_name': 'project_name',
        'author_name': 'DrivenData',
        'open_source_license': 'BSD-3-Clause',
        'python_interpreter': 'python3',
        'setup': True,
        }


def system_check(basename):
    platform = sys.platform
    if 'linux' in platform:
        basename = basename.lower()
    return basename


# Look into https://docs.pytest.org/en/latest/example/simple.html#request-example
# for handling command line options like skipping setup.
@pytest.fixture(scope='class', params=[{}, args])
def default_baked_project(tmpdir_factory, request):
    temp = tmpdir_factory.mktemp('data-project')
    out_dir = Path(temp).resolve()

    pytest.param = request.param
    # TODO - handle via a command-line option
    # pytest.param = {"setup": False}
    setup = True

    pytest.param = {
        "whisk_dependency": "-e {}".format(os.getcwd())
    }

    project_name = pytest.param.get('project_name') or 'project_name'
    proj_dir = create(project_name, output_dir=out_dir)
    proj = Path(proj_dir)
    if setup:
        whisk.setup.setup(proj)

    request.cls.path = proj
    yield

    # cleanup after
    shutil.rmtree(out_dir)
    # jupyter kernelspec list
    os.system("jupyter kernelspec uninstall {} -f".format(project_name))



import click
from whisk import whisk
from whisk.cli.commands.project.setup import cli as setup_cli
import logging

logger = logging.getLogger(__name__)

@click.command()
@click.argument('name')
@click.option('--setup/--no-setup', default=True, help="Run setup script after creating directory structure.")
@click.option('--force/--no-force', default=False, help="Overwrite the contents of output directory if it exists.")
@click.option('-o', '--output_dir', default=".", help="The parent directory that will contain the project.")
@click.pass_context
def cli(ctx, name, setup, output_dir, force):
    """
    Creates a Whisk project in the NAME directory.

    When --setup is true (the default), `whisk setup` is run after the project structure has been created
    to setup the environment.
    """
    res = whisk.create(name, output_dir=output_dir, force=force)
    if setup:
        ctx.invoke(setup_cli, dir = res)
    else:
        logger.info("Skipping setup. Run later with `whisk setup`.")
    return res



#!/usr/bin/env python
import pytest

# https://click.palletsprojects.com/en/7.x/testing/
from click.testing import CliRunner
from {{cookiecutter.project_name}}.cli.main import cli

def test_cli():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code == 0
    help_result = runner.invoke(cli, ['--help'])
    assert help_result.exit_code == 0

def test_predict_via_cli():
    """Test generating a model prediction via the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli, ['predict','[[1,2],[3,4]]'])
    assert result.exit_code == 0, "Predict failed with error: "+ result.output
    assert len(result.output) > 0



import whisk
import pickle

class Model:
    """
    This class should be used to load and invoke the serialized model and any other required
    model artifacts for pre/post-processing.
    """

    def __init__(self):
        """
        Load the model + required pre-processing artifacts from disk. Loading from disk is slow,
        so this is done in `__init__` rather than loading from disk on every call to `predict`.

        Tensorflow example:

            self.model = load_model(whisk.artifacts_dir / "model.h5")

        Pickle example:

            with open(whisk.artifacts_dir / 'tokenizer.pickle', 'rb') as file:
                self.tokenizer = pickle.load(file)
        """
        # REPLACE ME - add your loading logic
        with open(whisk.artifacts_dir / "model.pkl", 'rb') as file:
            self.model = pickle.load(file)

    def predict(self,data):
        """
        Returns model predictions.
        """
        # Add any required pre/post-processing steps here.
        return self.model.predict(data)



import os
import pytest
import json
# import app from the main.py file
from main import app

@pytest.fixture
def client():
    app.config['TESTING'] = True

    with app.test_client() as client:
        yield client

def test_predict(client):
    # The code below will likely fail when a real model
    # is provided. It should pass on a fresh app.
    response = client.post('/predict',
                            data=json.dumps({"data": [[1]]}),
                            content_type='application/json')
    assert response.status_code == 200



import click
import json
from {{cookiecutter.project_name}}.models.model import Model

@click.group()
def cli():
    pass

@cli.command()
@click.argument('x')
def predict(x):
    """
    Generate a model prediction.
    Example usage: {{cookiecutter.project_name}} predict [[1,2],[3,4]]
    """
    model = Model()
    x_parsed = json.loads(x)
    res = model.predict(x_parsed)
    click.echo(res)

if __name__ == "__main__":
    cli()



import click
import json
import boto3
import botocore
from whisk.project import Project

project = Project()
"""A reference to the associated :class:`whisk.project.Project` object for this project."""

PUBLIC_READ_ONLY_POLICY = {
  "Version":"2012-10-17",
  "Statement":[
    {
      "Sid":"PublicRead",
      "Effect":"Allow",
      "Principal": "*",
      "Action":["s3:GetObject"],
      "Resource":["arn:aws:s3:::[BUCKET_NAME]/*"]
    }
  ]
}
"""The public read-only policy to apply to an S3 bucket."""

DEFAULT_BUCKET_NAME = "whisk-" + project.name.replace("_","-")
"""The default bucket name for the project."""

S3_CLIENT = boto3.client('s3')

@click.group()
def cli():
    pass

@cli.command()
@click.option("--bucket", default=DEFAULT_BUCKET_NAME, show_default=True)
def create(bucket):
    """
    Creates an S3 bucket to store data. If no bucket name is provided, the bucket name is generated from the project directory name.
    """
    S3_CLIENT.create_bucket(Bucket=bucket)
    click.echo("{} S3 bucket created.".format(bucket))

@cli.command()
@click.option("--bucket", default=DEFAULT_BUCKET_NAME, show_default=True)
def delete(bucket):
    """
    Delete the S3 bucket used for data storage. If no bucket name is provided, the bucket name is generated from the project directory name.
    """
    bucket_resource = boto3.resource('s3').Bucket(bucket)
    try:
        bucket_resource.objects.all().delete()
    except Exception as e:
        click.echo("Unable to delete S3 bucket with name={}.\nError={}".format(bucket,e))
        exit(1)
    S3_CLIENT.delete_bucket(Bucket=bucket)
    click.echo("{} S3 bucket deleted.".format(bucket))

@cli.command()
@click.option("--bucket", default=DEFAULT_BUCKET_NAME, show_default=True)
def make_public(bucket):
    """
    Creates a public read-only policy for the S3 bucket.

    This is useful when end users need access to data or artifacts not stored in the code version control.

    Use ``whisk delete-policy`` to revert.
    """
    policy = PUBLIC_READ_ONLY_POLICY
    S3_CLIENT.put_bucket_policy(Bucket=bucket, Policy=json.dumps(policy).replace("[BUCKET_NAME]",bucket))
    click.echo("{} S3 bucket is now read-only to the public.".format(bucket))

@cli.command()
@click.option("--bucket", default=DEFAULT_BUCKET_NAME, show_default=True)
def delete_policy(bucket):
    """
    Deletes the policy associated with the S3 bucket.
    """
    S3_CLIENT.delete_bucket_policy(Bucket=bucket)
    click.echo("{} S3 bucket deleted.".format(bucket))



import flask
import whisk.dvc as dvc
from {{cookiecutter.project_name}}.models.model import Model
import sys
import os

# initialize the Flask application
app = flask.Flask(__name__)
# Pull the output of the DVC stage used to generate the serialized model when running on a
# deployed server. For example:
# dvc.pull("train.dvc")
model = Model()

@app.route("/predict", methods=["POST"])
def predict():
    """
    Accepts a payload w/content-type="application/json" and expects a `data` key/value.
    `data` is passed into the model for inference.
    """
    input = flask.request.json['data']
    result = model.predict(input)
    return flask.jsonify(result)

if __name__ == "__main__":
    app.run()



"""
Whisk includes default cli commands for performing common tasks in an ML project.
You can add new commands to your project and override existing commands. Files in this directory are
loaded as click commands.

This is an example of adding a new command (whisk app hello) and
overriding an existing command (whisk app start). Uncomment the code below to try it.

Click docs: https://click.palletsprojects.com/
For commands to load, the file must have a click `cli` command or group.
"""

import click
import subprocess

# @click.group()
# def cli():
#     pass
#
# @cli.command()
# def hello():
#     """Hello!"""
#     click.echo("Hello!")
#
# @cli.command()
# def start():
#     """Start override!!!!"""
#     click.echo("start override!!!!!!!!!!!!")



"""Main module."""
import whisk
from whisk.cli.log_tree import PARENT_TREE_NODE_PREFIX, CHILD_TREE_NODE_PREFIX
from cookiecutter.main import cookiecutter
import os
from os.path import dirname, realpath
# https://docs.python.org/3/library/pathlib.html
# Object-oriented filesystem paths
from pathlib import Path
import logging

logger = logging.getLogger(__name__)

def root_module_dir():
    """
    Returns a Path object with the root whisk module directory.
    """
    filepath = realpath(__file__)
    return Path(filepath).parents[0]

def cookiecutter_template_dir():
    return str(root_module_dir() / 'template/')

def project_name_to_slug(project_name):
    """
    Converts a raw project name to a slug:

    * Makes all letters lowercase
    * Replaces spaces with underscores
    """
    return project_name.lower().replace(' ', '_')

def create(project_name, output_dir=".", force=False):
    """
    Creates a whisk project.

    Parameters
    ----------
    project_name : str
        Name of the directory to create for the project. This is converted to a slug via :func:`project_name_to_slug`.

    output_dir : str, optional
        Path to create the directory. Default is the current working directory.

    force : bool, optional
        Recreates the project directory if it exists. Default is `False`.
    """
    # Locks to a specific version as earlier and later versions of whisk could expect a different
    # template structure.
    whisk_version = "whisk=={}".format(whisk.__version__)

    project_name_slug = project_name_to_slug(project_name)

    # `whisk_dependency` is more flexible (for example, specifying a local install)
    # than `whisk_install_requires` and is used in testing to require the local version of whisk.
    extra_content = {
        "project_name": project_name_slug,
        # Added to the project's requirements.txt
        "whisk_dependency": whisk_version,
        # Added to the project's setup.py file
        "whisk_install_requires": whisk_version
    }
    logger.debug("Creating whisk project with extra_content={}".format(extra_content))
    logger.info(PARENT_TREE_NODE_PREFIX+"Creating project directory structure...")
    res = cookiecutter(cookiecutter_template_dir(),
                no_input=True,
                overwrite_if_exists=force,
                output_dir=output_dir,
                extra_context=extra_content)
    logger.info(CHILD_TREE_NODE_PREFIX+"Project created in %s", res)
    logger.info(CHILD_TREE_NODE_PREFIX+"DONE.")
    return res



class ModelStub:
    # DELETE ME - Remove this class when you load a real model. This only
    # exists so the ModelWrapper works when creating an initial project.
    """
    A placeholder for a real ML Model. Returns the number of features in each row.

    Example:

        ModelStub().predict([[1,2],[3,4]]) => [2,2]
    """
    def predict(self,X):
        return list(map(lambda instance: len(instance), X))



"""Manage the web service."""
import click
from subprocess import call
import whisk.git as git

@click.group()
def cli():
    pass

@cli.command()
def start():
    """Start the HTTP web service."""
    call("honcho -f app/Procfile.dev start", shell=True)

@cli.command()
@click.argument("name")
def create(name):
    """
    Create a Heroku app for the web service with the given NAME. The NAME must be unique across all Heroku apps.

    A Heroku deploy performs the following steps:

    * Ensures there are no unstaged commits before creating the app. If there are uncommited changes the command exists with a warning.\n
    * Creates the Heroku app\n
    * Adds the `Multi-Procfile buildpack <https://github.com/heroku/heroku-buildpack-multi-procfile />`_ to access the Procfile within the ``app/`` folder of the project.\n
    * Adds the `Python buildpack <https://elements.heroku.com/buildpacks/heroku/heroku-buildpack-python />`_.\n
    * Sets a `PROCFILE` config var to ``app/Procfile``.\n
    * Pushes the git repo to Heroku.\n
    """
    if git.has_unstaged_changes():
        click.echo("This project has uncommitted changes.\n\nPlease add and commit the files to the Git repo, then retry:\n\ngit add .\ngit commit -m 'First Commit'\n")
        exit(1)
    call("heroku create -a {}".format(name), shell=True)
    call("heroku buildpacks:add -a {} https://github.com/heroku/heroku-buildpack-multi-procfile".format(name), shell=True)
    call("heroku buildpacks:add heroku/python", shell=True)
    call("heroku config:set PROCFILE=app/Procfile", shell=True)
    call("git push heroku master", shell=True)

@cli.command()
def destroy():
    """
    Delete the Heroku app.
    """
    call("heroku apps:destroy", shell=True)



"""
This module contains whisk command line interface (CLI) commands and logic for managing whisk projects.

Commands within :mod:`whisk.cli.commands.project` are only loaded when the current working directory is the top-level of a whisk project.

:class:`whisk.cli.whisk_multi_command.WhiskMultiCommand` loads the commands based on the current working directory. If ran within a project, it also loads commands defined in the ``whisk_commands/`` directory.
"""



from whisk.project import Project
import whisk.git as git
from whisk.cd import cd
from whisk.cli.log_tree import PARENT_TREE_NODE_PREFIX, CHILD_TREE_NODE_PREFIX
import os
import subprocess
from pathlib import Path
import logging
import sys
from subprocess import PIPE, STDOUT

logger = logging.getLogger(__name__)

NOTEBOOK_EXAMPLE_PATH = "notebooks/getting_started.ipynb"

def log_subprocess_output(logger_with_level,log_lines):
    """
    Calls the logger on each of the log_lines. The log lines are prefixed with :attr:`whisk.cli.log_tree.CHILD_TREE_NODE_PREFIX`.

    Parameters
    ----------
    logger_with_level : Logger
        A logger instance (ex: ``logger.debug``).

    log_lines : str
        A string of log lines separated with ``\\n``.
    """
    for line in log_lines.splitlines(): # b'\n'-separated lines
        logger_with_level(CHILD_TREE_NODE_PREFIX+line)

def exec(desc,cmd):
    """
    Executes the `cmd`, and logs `desc` prior to execution and "DONE" after.
    If the `cmd` has `stdout` or `stderr` output this is logged as well.

    If the exit code is nonzero, raises a ``SystemExit`` execption.

    Parameters
    ----------
    desc : str
        A description of the command operation.

    cmd : str
        The command to execute.
    """
    logger.info(PARENT_TREE_NODE_PREFIX+desc+"...")
    completed_process = None
    try:
        # universal_newlines is the same as `text`. text was added in 3.7:
        # Changed in version 3.7: Added the text parameter, as a more understandable alias of universal_newlines.
        # Added the capture_output parameter.
        completed_process = subprocess.run(cmd, shell=True, check=True, universal_newlines=True, stdout=PIPE, stderr=PIPE)
        log_subprocess_output(logger.info,completed_process.stdout)
        log_subprocess_output(logger.error,completed_process.stderr)
        logger.info(CHILD_TREE_NODE_PREFIX+"DONE.")
    except subprocess.CalledProcessError as e:
        logger.error(CHILD_TREE_NODE_PREFIX+"Error executing `%s`.\n       Stack trace:\n\n", cmd, exc_info=True)
        raise SystemExit("\n###\nSETUP FAILED. See whisk troubleshooting docs for help:\nhttps://whisk.readthedocs.io/en/latest/troubleshooting.html\n###\n")



def exec_setup(nbenv):
    exec("Setting up venv","python3 -m venv {}/venv".format(os.getcwd()))
    exec("Installing Python dependencies via pip  (may take several minutes)","venv/bin/pip install -r requirements.txt > /dev/null")
    logger.info(PARENT_TREE_NODE_PREFIX+"Initializing the Git repo")
    # Idempotent so just execute
    os.system("git init > /dev/null 2>&1")
    logger.info(CHILD_TREE_NODE_PREFIX+"DONE.")
    # Would rather use --sys-prefix, but not working:
    # https://github.com/jupyter/notebook/issues/4567
    exec("Setting up kernel={} for Jupyter Notebooks".format(nbenv),"venv/bin/python -m ipykernel install --user --name={}".format(nbenv))
    set_example_notebook_kernel(nbenv)
    # direnv will fail if not installed
    os.system("cp .envrc.example .envrc")
    os.system("direnv allow . > /dev/null 2>&1")
    if git.has_unstaged_changes():
        exec("Adding files to git", "git add .")
        exec("Making initial Git commit", "git commit -m 'Initial project structure' --author=\"Whisk <whisk@whisk-ml.org>\" > /dev/null")

def set_example_notebook_kernel(nbenv):
    """
    Updates the :attr:`NOTEBOOK_EXAMPLE_PATH` notebook kernel to use the kernel with name ``nbenv``.
    """
    nb_file = Path(NOTEBOOK_EXAMPLE_PATH)
    if not nb_file.is_file():
        logger.info("Getting started notebook does not exist @ {}. Not applying venv.".format(nb_file))
        return
    # Read in the file
    with open(NOTEBOOK_EXAMPLE_PATH, 'r') as file :
      filedata = file.read()

    # This could be run after the initial cookiecutter install.
    filedata = filedata.replace("{{cookiecutter.project_name}}", nbenv)

    # Write the file out again
    with open(NOTEBOOK_EXAMPLE_PATH, 'w') as file:
        file.write(filedata)

def log_success(dir):
    """Logs that the setup completed successfully and provides next steps."""
    logger.info("│\n└── Setup completed.")
    logger.info("\n###\nYOUR PROJECT IS READY. See the docs for help: https://whisk.readthedocs.io\n")
    logger.info("Get started:\n")
    if os.getcwd() != dir:
        # If not in the project directory, instruct the user to cd into the project.
        logger.info("cd %s", dir)
    logger.info("source venv/bin/activate")
    logger.info("###\n")

def log_pip_freeze():
    """Logs the output of ``pip freeze`` at the debug level."""
    reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'], universal_newlines=True)
    logger.debug("pip freeze output:\n"+reqs)

def setup(dir):
    """
    Sets up the project environment.

    Setup performs the following actions after changing the working directory to ``dir``:

    * Creates a `Python3 venv <https://docs.python.org/3/library/venv.html />`_ named "venv"
    * Installs the dependencies listed in the project's requirements.txt.
    * Initializes a Git repo
    * Creates an iPython kernel for use in Jupyter notebooks with name = <project_name>.
    * Creates a ``.envrc`` file based on ``.envrc.example`` for use with `direnv <https://github.com/direnv/direnv />`_.
      direnv loads environment variables listed in ``.envrc`` into the shell and is also used to auto-activate and
      deactivate the venv when entering and exiting the project directory.
    * Calls ``direnv allow .`` so the ``.envrc`` file can be loaded.
    * Makes an initial Git commit

    Parameters
    ----------
    dir : str
        The full path to the project directory.
    """
    project = Project(dir)
    project.validate_in_project()
    nbenv = project.name
    with cd(dir):
        logger.debug("cwd={}".format(os.getcwd()))
        exec_setup(nbenv)
    log_success(dir)
    log_pip_freeze()



from subprocess import check_output

def has_unstaged_changes():
    """Returns True if the git repo in the current directory has unstaged changes."""
    res=check_output("git status --porcelain",shell=True, universal_newlines=True)
    return ("\n" in res)



"""
Console script for whisk. The structure is based on `Click complex example app <https://github.com/pallets/click/blob/master/examples/complex/>`_.
"""
import sys
import click
import whisk
from whisk.cli.whisk_multi_command import WhiskMultiCommand
from whisk.log import configure_logger
import cookiecutter.log
import logging
import platform

def log_system_info(logger):
    """Logs information useful for debugging across all whisk commands."""
    logger.debug("whisk version=%s",whisk.__version__)
    logger.debug("whisk module location=%s",whisk)
    logger.debug("Platform: %s", platform.platform())
    logger.debug("Python Version: %s", platform.python_version())
    logger.debug("Python executable: %s", sys.executable)

@click.command(cls=WhiskMultiCommand)
@click.option('--debug/--no-debug', default=False, help="Enable verbose logging.")
@click.option('--log-file', default=None, help="Log output to a file.")
@click.pass_context
def cli(ctx, debug, log_file):
    """Entry point for the whisk cli."""
    # Ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)
    ctx.obj['DEBUG'] = debug
    ctx.obj['LOG_FILE'] = log_file
    configure_logger(stream_level='DEBUG' if debug else 'INFO', log_file=log_file)
    # Pass the same logging settings thru to cookiecutter.
    cookiecutter.log.configure_logger(stream_level='DEBUG' if debug else 'INFO', debug_file=log_file)
    logger = logging.getLogger(__name__)
    log_system_info(logger)

if __name__ == "__main__":
    cli()



import click
from whisk import whisk
from whisk.project import Project
import os

class WhiskMultiCommand(click.MultiCommand):
    """
    This is a custom class based on https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
    that loads click command files at run time from both the whisk module commands and commands
    specified within the created project. This lets users override existing whisk cli commands and create new commands.

    Project-specific commands are only loaded if the the cwd is the project root.
    """

    def core_commands_dir(self):
        """Commands available whether in or outside a project."""
        return whisk.root_module_dir() / "cli/commands"

    def core_project_commands_dir(self):
        """whisk-provided project-only commands directory."""
        return self.core_commands_dir() / "project"

    def project_commands_dir(self):
        """Custom project commands directory."""
        return Project().commands_dir

    def list_commands(self, ctx):
        """
        MultiCommand classes must implement this function.
        This returns a set of file names across all directories.
        """
        project = Project()
        rv = []
        rv = rv + self._command_file_names(self.core_commands_dir())
        if project.in_project():
            rv = rv + self._command_file_names(self.core_project_commands_dir())
            # Possible commands directory is deleted in project
            if project.commands_dir.exists():
                rv = rv + self._command_file_names(project.commands_dir)

        rv.sort()
        return set(rv)

    def _command_file_names(self, dir_name):
        """
        Returns a list of file names minus the *.py file extension.
        This assumes that click commands are listed in directories containing only click command files.
        """
        rv = []
        for filename in os.listdir(dir_name):
            if filename != '__init__.py' and filename.endswith(".py"):
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def _eval_file(self,fn):
        """
        Evals the given filename, return the 'cli' namespace.
        The file must contain a 'cli' click command or group.

        If the file doesn't exist, nothing is returned.
        """
        if not os.path.isfile(fn):
            return

        ns = {}
        with open(fn) as f:
            code = compile(f.read(), fn, 'exec')
            eval(code, ns, ns)

        if 'cli' in ns:
            return ns['cli']
        else:
            # The code was eval'd but did not have a `cli` key in the namespace.
            return

    def get_command(self, ctx, name):
        """
        MultiCommand classes must implement this function.

        Returns the CLI command w/the given name. The following load order is used:
        * Project-specific commands
        * Core commands
        * Core project-specific commands

        If a command is found it is returned immediately. For example, if both a custom project-specific
        command and a core project-specific command exist with the same name the custom command is returned,
        overriding the default.
        """
        project = Project()

        if project.in_project():
            fn = os.path.join(self.project_commands_dir(), name + '.py')
            ns = self._eval_file(fn)

            if ns:
                return ns

        fn = os.path.join(self.core_commands_dir(), name + '.py')
        ns = self._eval_file(fn)

        if ns:
            return ns

        fn = os.path.join(self.core_project_commands_dir(), name + '.py')
        ns = self._eval_file(fn)
        return ns



import click
from subprocess import call, check_output
from whisk.project import Project

project = Project()

@click.group()
def cli():
    pass

@cli.command()
def dist():
    """Builds a source distribution of the project."""
    call("python setup.py sdist > /dev/null", shell=True)
    click.echo("Python Package created in {}:".format(project.path / "dist"))
    res = check_output("ls -h dist", shell=True)
    click.echo(res)



import os
import pytest
import datetime as dt
from subprocess import check_output, CalledProcessError
from conftest import system_check
from pathlib import Path


def no_curlies(filepath):
    """ Utility to make sure no curly braces appear in a file.
        That is, was Jinja able to render everything?
    """
    with open(filepath, 'r') as f:
        data = f.read()

    template_strings = [
        '{{',
        '}}',
        '{%',
        '%}'
    ]

    template_strings_in_file = [s in data for s in template_strings]
    return not any(template_strings_in_file)

def changed_files(root_path,ago):
    """
    Returns a list of files changed within the `root_path` since `ago`.
    """
    now = dt.datetime.now()
    changed_files = []
    for root, dirs,files in os.walk(root_path):
        for fname in files:
            path = os.path.join(root, fname)
            st = os.stat(path)
            mtime = dt.datetime.fromtimestamp(st.st_mtime)
            if mtime > ago:
                changed_files.append(path)
    return changed_files

def paths_to_relative(root,paths):
    """
    Converts a list of absolute paths to paths relative to the the root project path.
    """
    return list(map(lambda x: str(Path(x).relative_to(root)), paths))

@pytest.mark.usefixtures("default_baked_project")
class TestCookieSetup(object):
    def test_project_name(self):
        project = self.path
        if pytest.param.get('project_name'):
            assert project.name == pytest.param.get('project_name')
        else:
            assert project.name == 'project_name'

    def test_readme(self):
        readme_path = self.path / 'README.md'
        assert readme_path.exists()
        assert no_curlies(readme_path)
        if pytest.param.get('project_name'):
            with open(readme_path) as fin:
                assert pytest.param.get('project_name') in next(fin).strip()

    def test_setup(self):
        setup_ = self.path / 'setup.py'
        args = ['python', setup_, '--version']
        p = check_output(args).decode('ascii').strip()
        assert p == '0.1.0'

    def test_requirements(self):
        reqs_path = self.path / 'requirements.txt'
        assert reqs_path.exists()
        assert no_curlies(reqs_path)
        if pytest.param.get('python_interpreter') == 'python':
            with open(reqs_path) as fin:
                lines = list(map(lambda x: x.strip(), fin.readlines()))
            assert 'pathlib2' in lines

    # def  test_app_starts(self):
        # Fails:
        # ModuleNotFoundError: No module named 'flask'
        # check_output(['venv/bin/inv', 'app.test'], cwd=self.path)

    def test_idempotent_install_command(self):
        # Running setup is slow, so by default setup=False.
        if not pytest.param.get("setup"):
            return

        ago = dt.datetime.now()
        check_output(['whisk setup'], cwd=self.path)
        changed = changed_files(self.path,ago)
        # Script appears to update some files but unsure if that changes any functionality.
        # assert len(changed) == 0, "Files were modified: {}".format(changed)

    def test_notebook_runs(self):
        # Running setup is slow, so by default setup=False.
        if not pytest.param.get("setup"):
            return
        check_output(["venv/bin/inv", "notebooks.run", "notebooks/getting_started.ipynb"],
                     cwd=self.path)

    def test_app_create(self):
        if not pytest.param.get("setup"):
            return
        with pytest.raises(CalledProcessError):
            # Ensure we can run this task, but it fails because of uncommitted changes.
            check_output(["venv/bin/inv", "app.create", "APPNAME"],
                         cwd=self.path)

    def test_model_predict(self):
        # Running setup is slow, so by default setup=False.
        if not pytest.param.get("setup"):
            return
        check_output(["venv/bin/inv", "model.predict", "[[1,2],[3,4]]"],
                     cwd=self.path)

    def test_folders(self):
        project = self.path
        project_name = project.name
        expected_dirs = [
            '.whisk',
            'app',
            'data',
            'data/external',
            'data/interim',
            'data/processed',
            'data/raw',
            'docs',
            'notebooks',
            'references',
            'reports',
            'reports/figures',
            'src',
            "src/{}".format(project_name),
            "src/{}/artifacts".format(project_name),
            "src/{}/data".format(project_name),
            "src/{}/features".format(project_name),
            "src/{}/models".format(project_name),
            "src/{}/visualization".format(project_name),
        ]

        # venv dir created if setup=True
        if pytest.param.get("setup"):
            expected_dirs.append("venv")

        ignored_dirs = [
            str(self.path)
        ]

        abs_expected_dirs = [str(self.path / d) for d in expected_dirs]
        abs_dirs, _, _ = list(zip(*os.walk(self.path)))

        rel_dirs = paths_to_relative(self.path,abs_dirs)
        expected_dirs = paths_to_relative(self.path,set(abs_expected_dirs + ignored_dirs))

        missing_dirs = set(expected_dirs) - set(rel_dirs)
        assert len(missing_dirs)  == 0, "Expected dirs are missing: {}".format(missing_dirs)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = ['Click>=7.0', 'cookiecutter>=1.7.2']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Derek Haynes",
    author_email='derek@dlite.cc',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Machine Learning Project Framework Generator",
    entry_points={
        'console_scripts': [
            'whisk=whisk.cli.cli:cli',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='whisk',
    name='whisk',
    packages=find_packages(include=['whisk', 'whisk.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/bookletai/whisk',
    version='0.1.30',
    zip_safe=False,
)



import click
import sys
import subprocess
from whisk.project import Project

@click.group()
def cli():
    pass

@cli.command()
@click.argument('relative_path')
def run(relative_path):
    """
    Run a notebook from the command line with the given RELATIVE_PATH.
    """
    project = Project()
    subprocess.call("jupyter nbconvert --clear-output --ExecutePreprocessor.timeout=1000 --execute {}".format(relative_path),shell=True)



"""Top-level package for whisk."""

__author__ = """Derek Haynes"""
__email__ = 'derek@dlite.cc'
__version__ = '0.1.30'

project = None
"""
Reference to the current :class:`whisk.project.Project`. Set when whisk is loaded from
within a created project.
"""

data_dir = None
"""
The location of the data directory as a `pathlib.Path <https://docs.python.org/3.8/library/pathlib.html#basic-use/>`_. This is set from the Project instance and made
available as a top-level attribute since it is frequently used.
"""

artifacts_dir = None
"""
The location of the artifacts directory as a `pathlib.Path <https://docs.python.org/3.8/library/pathlib.html#basic-use/>`_. This is set from the Project instance and made
available as a top-level attribute since it is frequently used.
"""



import os
from subprocess import check_output

def pull(dvc_file):
    """
    Pulls the output of the specified `dvc_file` into the repository.
    This is useful when running outside the local environment (like a deployed web server)
    that doesn't have all data files by default.

    If `dvc pull` fails (has a non-zero exit status), a `subprocess.CalledProcessError` exception
    is raised.
    """
    # The deployed app isn't a git repo but needs to be for dvc.
    # `git init` is idempotent.
    os.system("git init")
    # Pull the training output (the serialized model) when running on a deployed server.
    check_output(["dvc", "pull", dvc_file])



from setuptools import find_packages, setup

setup(
    name='{{cookiecutter.project_name}}',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    version='0.1.0',
    include_package_data=True,
    # By default only whisk is added as a dependency. whisk is required for
    # accessing whisk.project.artifacts_dir in the default models.model.Model class.
    # You likely need to list more dependencies for your model package.
    # For example: your model framework (Scikit, Torch, etc), numpy, Pandas, etc.
    #
    # Copying the requirements.txt dependencies into `install_requires=` isn't recommended as this
    # includes many dependencies that are not required to generate predictions and can result in a very large package.
    #
    # You can test that the package works and contains needed dependencies by running `tox` from the
    # command line. tox tests the package in an isoloated venv.
    install_requires=[
        '{{cookiecutter.whisk_install_requires}}'
    ],
    entry_points={
        'console_scripts': [
            '{{cookiecutter.project_name}}={{cookiecutter.project_name}}.cli.main:cli',
        ],
    },
    description='A short description of the project.',
    author='Your name (or your organization/company/team)',
    author_email='you@example.com',
    python_requires='>=3.6',
    url="https://ADD THE URL TO YOUR PROJECT GITHUB REPO OR DOCS"
)



from os.path import realpath
import whisk
from whisk.project import Project

"""
Initializes the whisk.project attribute so project helpers can be accessed via
whisk.project.*.
"""
whisk.project = Project.from_module(realpath(__file__))
whisk.data_dir = whisk.project.data_dir
whisk.artifacts_dir = whisk.project.artifacts_dir



import os
# https://docs.python.org/3/library/pathlib.html
# Object-oriented filesystem paths
from pathlib import Path

class Project:
    """
    Abstracts project-level attributes into a class.
    """

    @classmethod
    def from_module(cls,path):
        """
        Takes a path from a generated project's src/{{name}}/__init__.py file
        and returns a Project object Initialized w/that path.
        """
        path_obj = Path(path)
        obj = cls(path_obj.parents[2],module_dir=path_obj.parents[0])
        return obj

    def __init__(self, dir=os.getcwd(), module_dir=None):
        """
        Initializes a project from the existing whisk project directory `dir`.
        """
        self.dir = dir
        """The top-level project directory (str)"""
        self.path = Path(self.dir)
        """The top-level project directory as a pathlib.Path"""
        self.name = self.name_from_src_dir(self.path / "src")
        """Name of the project derived from the src/<project_name> directory (str)"""
        self.module_dir = module_dir
        """
        Location of the project's module directory as a pathlib.Path.
        This is derived from `path` if not provided.
        """
        if not self.module_dir:
            self.module_dir = self.path / "src/{}".format(self.name)
        self.artifacts_dir = self.module_dir / "artifacts"
        """
        Location of the project's artifacts directory as a pathlib.Path.
        This is derived from `path` if not provided.
        """
        self.data_dir = None
        self.commands_dir = None
        if self.in_project():
            self.data_dir = self.path / "data"
            """
            Location of the project's data directory as a pathlib.Path.
            Returns `None` if not within a whisk project.
            """
            self.commands_dir = self.path / "whisk_commands"
            """
            Location of the project's whisk commands directory as a pathlib.Path.
            Returns `None` if not within a whisk project.
            """

    def name_from_src_dir(self, src_dir):
        """
        Derives the project name from the first child directory in the ``src/``
        directory that contains an ``__init__.py`` file.

        Parameters
        ----------
        src_dir : pathlib.Path
            Path object referencing the project ``src/`` directory.
        """

        src_children = sorted(src_dir.glob("*/__init__.py"))
        if len(src_children) == 0:
            return None
        return src_children[0].parent.stem

    def validate_in_project(self):
        """
        Raises an `OSError` if this is not a whisk project.
        """
        if not self.in_project():
            raise OSError("{} is not a whisk project directory.".format(self.dir))

    def in_project(self):
        """
        Returns True if the project's path is the root whisk
        project directory.

        This is deteremined by the presence of the ".whisk/" directory.
        """
        return (self.path / ".whisk").is_dir()
